<?php

class Table{
	
	/**
	 * Contient le nom des colonnes de la table.
	 * Exemple : array('id', 'firstName', 'lastName', 'password')
	 **/
	private $headers = array();
	
	/**
	 * Contient les données de la table, à savoir un tableau normalement à 2 dimensions, à 3 dimensions en cas de regroupements (GROUP BY).
	 * Exemple : 
	 * array(
	 *		array(1, 'Jérémy', 'Dupond', '123456'),
	 *		array(2, 'Wilfried', 'Kassovick', '654321'),
	 *		...
	 * )
	 **/
	private $data = array();
	
	/**
	 * Construit un table: en-têtes et données.
	 * @param Array $headers
	 *		Nom des colonnes (en-têtes).
	 * @param Array $data
	 *		Données de la tables. Les lignes de la tables doivents comporter autant de colonnes qu'il y a de colonnes d'en-tête. 
	 *		Dans le cas contraire, une exception est levée.
	 **/
	public function __construct($headers, $data){
		// Vérifications
		if(!is_array($headers) || !is_array($data)){
			throw new Exception('Appel au constructeur Table::__construct($header, $data) invalide. Attendu : paramètres $headers et $data de type tableau.');
		}
		$n = count($headers);
		foreach($data as $i => $row){
			if(!is_array($row)){
				throw new Exception('Appel au constructeur Table::__construct($header, $data) invalide. Les données du tûple ' . $i . ' sont invalides. Attendu : tableau.');
			}
			if(count($row) != $n){
				throw new Exception('Appel au constructeur Table::__construct($header, $data) invalide. Les données du tûple ' . $i . ' sont invalides. Attendu : tableau à ' . $n . ' colonnes.');
			}
		}
		// Si tout est OK
		$this->headers = $headers;
		$this->data = $data;
	}
	
	/**
	 * Selon le cas, retourne le nom ou l'index de la colonne. Retourne null si la colonne n'existe pas.
	 * @param Mixed $index
	 * 		Deux fonctionnements: 
	 * 		- $table->getHeader(5) retourne le nom de la 6ème colonne (6ème en-tête) ;
	 * 		- $table->getHeader('firstName') retourne l'index de la colonne dont le nom est 'firstName'
	 * @return Mixed
	 **/
	public function getHeader($index){
		if(is_int($index)){
			if(isset($this->headers[$index])){
				return $this->headers[$index];
			}
			throw new Exception('Colonne ' . $index . ' (int) inconnue.');
		}else if(is_string($index)){
			foreach($this->headers as $i => $header){
				if($header == $index){
					return $i;
				}
			}
			throw new Exception('Colonne "' . $index . '" (string) inconnue.');
		}
		throw new Exception('Appel de Table::getHeader($index) incorrect. Attendu : paramète $index de type int ou string.');
	}
	
	/**
	 * Retourne le nom des colonnes
	 * @return Array
	 **/
	public function getHeaders(){
		return $this->headers;
	}
	
	/**
	 * Retourne les données de la table
	 * @return Array
	 **/
	public function getData(){
		return $this->data;
	}
	
	/**
	 * Retourne la table résultant du produit cartésien entre la table courante et la table passée en paramètre.
	 * @param Table $table
	 * @return Table
	 *		
	 **/
	public function product(Table $table){
		$newHeaders = array_merge($this->headers, $table->getHeaders()); // array_merge concatène les valeurs de 2 tableaux pour former un tableau unique
		$newData = array();
		$data = $table->getData();
		foreach($this->data as $leftRow){
			foreach($data as $rightRow){
				$newData[] = array_merge($leftRow, $rightRow);
			}
		}
		return new Table(
			$newHeaders,
			$newData
		);
	}
	
	/**
	 * Retourne la table résultat de la jointure entre la table courante et la table passée en paramètre.
	 * Autrement dit, il s'agit d'implémenter: $this T1 INNER JOIN $table T2 ON T1.$field1 = T2.$field2
	 * @param Table $table
	 *		Table avec laquelle on fait la jointure.
	 * @param Mixed $field1
	 *		Nom de la colonne de la table courante sur laquelle opérer la jointure
	 * @param Mixed $field2
	 *		Nom de la colonne de $table sur laquelle opérer la jointure.
	 * @return Table
	 **/
	public function innerJoin(Table $table, $field1, $field2){
		return $this->product($table)->where($field1, '=', $field2);
	}
	
	/**
	 * Retourne la table résultant de la jointure externe à gauche effectuée entre la table courante et la table passée en 
	 * paramètre. 
	 * Autrement dit, il s'agit d'implémenter: $this T1 LEFT JOIN $table ON T1.$field1 = T2.$field2.
	 * @param Table $table
	 *		Table avec laquelle on fait la jointure.
	 * @param Mixed $field1
	 *		Nom de la colonne de la table courante sur laquelle opérer la jointure
	 * @param Mixed $field2
	 *		Nom de la colonne de $table sur laquelle opérer la jointure.
	 * @return Table
	 **/
	public function leftJoin(Table $table, $field1, $field2){
		// finalement, lorsqu'on a une jointure externe à gauche, il s'agit de :
		// - récupérer toutes lignes de la table de gauche 
		// - de trouver la correspondance avec les lignes de la table de droite ;
		// - s'il n'y a pas de correspondance, on prend la ligne de gauche et on complète par des null.
		// traduction :
		$newHeaders = array_merge($this->getHeaders(), $table->getHeaders());
		$newData = array();
		$left = $this->getData();
		$right = $table->getData();
		$indexLeft = $this->getHeader($field1);
		$indexRight = $table->getHeader($field2);
		$countHeaders = count($newHeaders);
		foreach($left as $rowLeft){
			$found = false;
			foreach($right as $rowRight){
				if($rowLeft[$indexLeft] === $rowRight[$indexRight]){
					$newData[] = array_merge($rowLeft, $rowRight);
					$found = true;
				}
			}
			if(!$found){
				$i = count($rowLeft);
				while($i < $countHeaders){
					$rowLeft[] = null;
					$i++;
				}
				$newData[] = $rowLeft;
			}
		}
		return new Table($newHeaders, $newData);
	}
	
	/**
	 * Retourne la table résultant de la jointure externe externe à doite effectuée entre la table courante et la table passée 
	 * en paramètre.
	 * Autrement dit, il s'agit d'implémenter: $this T1 RIGHT JOIN $table ON T1.$field1 = T2.$field2.
	 * @param Table $table
	 *		Table avec laquelle on fait la jointure.
	 * @param Mixed $field1
	 *		Nom de la colonne de la table courante sur laquelle opérer la jointure
	 * @param Mixed $field2
	 *		Nom de la colonne de $table sur laquelle opérer la jointure.
	 * @return Table
	 **/
	public function rightJoin(Table $table, $field1, $field2){
		// finalement, un $this T1 RIGHT JOIN $table T2 ON T1.$field1 = T2.$field2 n'est rien d'autre qu'un 
		// $table T1 RIGHT JOIN $this T2 ON T1.$field2 = T2.$field1 
		return $table->leftJoin($this, $field2, $field1);
	}
	
	/**
	 * Retourne la table résultat d'une restriction. Autrement dit, il s'agit d'implémenter une clause WHERE sql.
	 * @param Table $field
	 *		Champ de la table courante.
	 * @param String $operator
	 *		Opérateur, typiquement "=" et "<>". Libre à vous de fournir ensuite une implémentation pour d'autres opérateurs (exemples: IN, LIKE, etc.).
	 * @param Mixed $value
	 *		Valeur(s) à laquelle on compare la valeur du champ. Attention ! La valeur $valeur peut être un autre champ de la table. 
	 *		Si $valeur est une chaîne, c'est qu'il s'agit soit d'un autre champ, soit d'une chaîne de caractères :
	 *		- Pour comparer le champ $field à une chaîne de caractère, la chaîne doit être placée entre des guillemets simples ou doubles.
	 *		  Exemple : $value = "'une chaîne'" ou $value = '"une chaîne"'
	 *		- Pour comparer le champ $field à un autre champ, le nom du champ est placé sans guillements.
	 *		  Exemple : $value= 'categoryId' ou $value = "categoryId".
	 * @return Table
	 **/
	public function where($field, $operator, $value){
		if(!is_string($field)){
			throw new Exception('Appel de Table::where(...) incorrect. Attendu : nom de colonne $field (string).');
		}
		$index = $this->getHeader($field);
		$newData = array();
		// vérifie si la comparaison porte sur un champ => à savoir, pour que $value soit considéré comme une chaîne et non un champ, il faut que la valeur 
		// contenue dans $value soit encadrée par des guillemets simple ou doubles. Exemples : $value = '"Ordinateur"' ou $value = "'Ordinateur'".
		if(is_string($value)){
			if($value[0] == "'" || $value[1] == '"'){
				// $value représente bien une chaîne
				$value = substr($value, 1, strlen($value)-2);
				$field = -1;
			}else{
				// $value représente un champ 
				$field = $this->getHeader($value);
			}
		}else{
			$field = -1;
		}
		// On ne conserve que les lignes pour lesquelles la comparaison est vérifiée
		foreach($this->data as $row){
			if($field != -1){
				$value = $row[$field];
			}
			if($operator == '='){
				if($row[$index] === $value){
					$newData[] = $row;
				}
			}else if($operator == '<>'){
				if($row[$index] !== $value){
					$newData[] = $row;
				}
			}else if($operator == '>'){
				if($row[$index] > $value){
					$newData[] = $row;
				}
			}else if($operator == '>='){
				if($row[$index] >= $value){
					$newData[] = $row;
				}
			}else if($operator == '<'){
				if($row[$index] < $value){
					$newData[] = $row;
				}
			}else if($operator == '<='){
				if($row[$index] <= $value){
					$newData[] = $row;
				}
			}else if($operator == 'IN'){
				if(in_array($row[$index], $value)){
					$newData[] = $row;
				}
			}else if($operator == 'LIKE'){
				if(preg_match('/' . str_replace('%', '.*', $value) . '/', $row[$index])){
					$newData[] = $row;
				}
			}else{
				throw new Exception('Appel de Table::where(...) incorrect. Attendu : opérateur $operator valide (=, <>, >, IN, LIKE, etc.).');
			}
		}
		return new Table($this->headers, $newData);
	}
	
	/**
	 * Retourne la table avec les seules colonnes demandées (on dit qu'il s'agit d'une projection). Autrement dit, il s'agit de fournir une implémentation 
	 * de la clause SELECT.
	 * @param Array $fields
	 *		Nom des colonnes à conserver. Evolution: gérer l'agrégat SUM('nomChamp') (éventuellement AVG('nomChamp') également).
	 * @return Table
	 **/
	public function select($fields){
		// on construit une tableau (associatif) contenant la liste des colonnes "sélectionnées". En clef : l'index du champ ; en valeur : true.
		$indexes = array();
		foreach($fields as $field){
			if(!is_string($field)){
				throw new Exception('Appel de Table::select(...) incorrect. Attendu : nom de colonnes $fields (array of strings).');
			}
			// cas des agrégats : COUNT, SUM et AVG
			if(substr($field, 0, 5) == 'COUNT'){
				$field = substr($field, 6, strlen($field)-7);
				$indexes[$this->getHeader($field)] = 'COUNT';
			}else if(substr($field, 0, 3) == 'SUM'){
				$field = substr($field, 4, strlen($field)-5);
				$indexes[$this->getHeader($field)] = 'SUM';
			}else if(substr($field, 0, 3) == 'AVG'){
				$field = substr($field, 4, strlen($field)-5);
				$indexes[$this->getHeader($field)] = 'AVG';
			}else{
				// cas classique: "sélection" (projection) d'une simple champ
				$indexes[$this->getHeader($field)] = true;
			}
		}
		$newHeaders = array();
		foreach($this->headers as $i => $header){
			if(isset($indexes[$i])){
				// cas des agrégats : COUTN, SUM et AVG
				if($indexes[$i] === 'COUNT' || $indexes[$i] === 'SUM' || $indexes[$i] === 'AVG'){
					$newHeaders[] = $indexes[$i] . '(' . $header . ')';
				}else{
					$newHeaders[] = $header;
				}
			}
		}
		$newData = array();
		foreach($this->data as $row){
			$newRow = array();
			foreach($row as $i => $column){
				if(isset($indexes[$i])){
					// cas des agrégats : COUNT, SUM et AVG
					if($indexes[$i] === 'COUNT'){
						if(!is_array($column)){
							throw new Exception('Appel de Table::select(...) incorrect. Attendu : agrégat COUNT(...) utilisé sur un champ agrégeable.');
						}
						$newRow[] = count($column);
					}else if($indexes[$i] === 'SUM'){
						if(!is_array($column)){
							throw new Exception('Appel de Table::select(...) incorrect. Attendu : agrégat SUM(...) utilisé sur un champ agrégeable.');
						}
						$sum = 0;
						foreach($column as $value){
							if(!is_numeric($value)){
								throw new Exception('Appel de Table::select(...) incorrect. Attendu : agrégat SUM(...) utilisé sur un champ numérique.');
							}
							$sum += $value;
						}
						$newRow[] = $sum;
					}else if($indexes[$i] === 'AVG'){
						if(!is_array($column)){
							throw new Exception('Appel de Table::select(...) incorrect. Attendu : agrégat AVG(...) utilisé sur un champ agrégeable.');
						}
						$sum = 0;
						foreach($column as $value){
							if(!is_numeric($value)){
								throw new Exception('Appel de Table::select(...) incorrect. Attendu : agrégat AVG(...) utilisé sur un champ numérique.');
							}
							$sum += $value;
						}
						$newRow[] = $sum / count($column);
					}else{
						$newRow[] = $column;
					}
				}
			}
			$newData[] = $newRow;
		}
		return new Table($newHeaders, $newData);
	}
	
	/**
	 * Il s'agit ici d'implémenter une clause GROUP BY. Mais que fait un GROUP BY ? Nous y reviendrons.
	 * @param Array $fields
	 *		Nom des colonnes sur lesquelles effectuer les regroupements.
	 * @return Table
	 **/
	public function groupBy($fields){
		// Une idée (si la méthode orderBy est implémentée) consiste à déjà trier les données.
		$sort = array();
		foreach($fields as $i => $field){
			$sort[] = $field . ' ASC';
		}
		$sort = implode(',', $sort);
		// Trions les données
		$table = $this->orderBy($sort);
		// Permettons nous de conserver l'index des champs plutôt que leur nom (vu que c'est le nom des champ qu'on nous a normalement fourni en paramètre)
		$indexes = array();
		foreach($fields as $i => $field){
			$index = $this->getHeader($field);
			$indexes[$index] = $index;
		}
		// Maintenant que les données sont triées, regroupons-les
		$data = $table->getData();
		$n = count($data);
		$newData = array();
		$newRow = array(); // regroupement en cours de construction
		if($n>0){
			$newRow = $data[0];
		}
		$group = true;
		for($i=1 ; $i<$n ; $i++){
			$curRow = $data[$i];
			$group = true;
			foreach($indexes as $index){
				// si les lignes ne sont pas regroupables, c'est que notre regroupement en cours de construction a fini d'être complété
				if($curRow[$index] != $newRow[$index]){
					$group = false;
					break;
				}
			}
			if($group){
				// on "ajoute la ligne" au regroupement
				foreach($curRow as $j => $curRowColumn){
					if(!isset($indexes[$j])){
						if(!is_array($newRow[$j])){
							$newRow[$j] = array($newRow[$j]);
						}
						$newRow[$j][] = $curRowColumn;
					}
				}
			}else{
				// regroupement terminé
				$newData[] = $newRow;
				// place au nouveau regroupement
				$newRow = $curRow;
			}
		}
		if($group){
			$newData[] = $newRow;
		}
		return new Table($this->headers, $newData);
	}
	
	/**
	 * Si vous souhaitez fournir une implémentation de la clause orderBy (utiliser les fonctions de tri PHP). Le cas échéant, choisissez d'autres paramètres.
	 * @param String $sort
	 *		Chaîne de la forme: 'champ1 ASC, champ2 DESC, ...'.
	 * @return Table
	 **/
	public function orderBy($sort){
		$newData = $this->data;
		// on converit une chaîne de la forme 'champ1 ASC, champ2 DESC, ...' en array(array(indexChamp1, 'ASC'), array(indexChamp2, 'DESC'), ...)
		$sort = explode(',', $sort);
		foreach($sort as $i => $criteria){
			$sort[$i] = explode(' ', $criteria);
			$sort[$i][0] = $this->getHeader($sort[$i][0]);
		}
		// la fonction usort permet de trier les éléments de $newData (paramètre en entrée/sortie).
		// Elle prend en paramètre :
		// - la tableau dont les élements sont à trier ;
		// - une fonction de comparaison. Cette fonction doit permettre de comparer deux éléments du tableau. Si le premier élément est inférieur strictement
		//   au second, la fonction doit retourner un nombre <0. A l'inverse si le premier élément est supérieur strictement au second, la fonction doit retourner 
		//   un nombre >0. Si les éléments sont égaux, la fonction doit typiquement retourner 0.
		usort($newData, function($row1, $row2) use($sort){
			foreach($sort as $criteria){
				$index = $criteria[0];
				$direction = $criteria[1];
				if($row1[$index] < $row2[$index]){
					if($direction == 'DESC'){
						return 1;
					}else{
						return -1;
					}
				}else if($row1[$index] > $row2[$index]){
					if($direction == 'DESC'){
						return -1;
					}else{
						return 1;
					}
				}
			}
			return 0;
		});
		return new Table($this->headers, $newData);
	}
	
	/**
	 * Ajoute une ligne dans la table courant. Autrement dit, il s'agit d'implémenter la clause sql INSERT INTO.
	 * @param Array $values
	 *		Valeurs (ligne) à insérer dans la table. Il doit y avoir une valeur par colonne (sinon Exception!). 
	 * @return $this
	 **/
	public function insert($values){
		// on vérifie que le nombre de valeur correspond bien au nombre de headers (i.e. nombre de champs)
		if(count($values) != count($this->headers)){
			throw new Exception('Appel de Table::insert(...) incorrect. Attendu : la ligne à insérer ne comporte pas le bon nombre de champs.');
		}
		// rien de plus simple
		$this->data[] = $values;
		return $this;
	}
	
	/**
	 * Supprime 0 ou plusieurs ligne de la table courante en fonction d'une condition. Autrement dit, il s'agit 
	 * d'implémenter: DELETE FROM $this WHERE $field $operator $value
	  * @param Table $field
	 *		Champ de la table courante.
	 * @param String $operator
	 *		Opérateur, typiquement "=" et "<>". Libre à vous de fournir ensuite une implémentation pour d'autres opérateurs (exemples: IN, LIKE, etc.).
	 * @param Mixed $value
	 *		Valeur(s) à laquelle on compare la valeur du champ. Attention ! La valeur $valeur peut être un autre champ de la table. 
	 *		Si $valeur est une chaîne, c'est qu'il s'agit soit d'un autre champ, soit d'une chaîne de caractères :
	 *		- Pour comparer le champ $field à une chaîne de caractère, la chaîne doit être placée entre des guillemets simples ou doubles.
	 *		  Exemple : $value = "'une chaîne'" ou $value = '"une chaîne"'
	 *		- Pour comparer le champ $field à un autre champ, le nom du champ est placé sans guillements.
	 *		  Exemple : $value= 'categoryId' ou $value = "categoryId".
	 * @return $this
	 **/
	public function delete($field, $operator, $value){
		// finalement, pour le DELETE FROM ... WHERE ..., on peut très largement s'inspirer de la clause WHERE.
		// Sauf que : il ne s'agit pas ici de conserver les lignes mais de les supprimer
		if(!is_string($field)){
			throw new Exception('Appel de Table::delete(...) incorrect. Attendu : nom de colonne $field (string).');
		}
		$index = $this->getHeader($field);
		$newData = array();
		// vérifie si la comparaison porte sur un champ => à savoir, pour que $value soit considéré comme une chaîne et non un champ, il faut que la valeur 
		// contenue dans $value soit encadrée par des guillemets simple ou doubles. Exemples : $value = '"Ordinateur"' ou $value = "'Ordinateur'".
		if(is_string($value)){
			if($value[0] == "'" || $value[1] == '"'){
				// $value représente bien une chaîne
				$value = substr($value, 1, strlen($value)-2);
				$field = -1;
			}else{
				// $value représente un champ 
				$field = $this->getHeader($value);
			}
		}else{
			$field = -1;
		}
		// On conserve tout sauf les lignes pour lesquelles la comparaison est vérifiée
		foreach($this->data as $row){
			if($field != -1){
				$value = $row[$field];
			}
			if($operator == '='){
				if($row[$index] !== $value){
					$newData[] = $row;
				}
			}else if($operator == '<>'){
				if($row[$index] === $value){
					$newData[] = $row;
				}
			}else if($operator == '>'){
				if($row[$index] <= $value){
					$newData[] = $row;
				}
			}else if($operator == '>='){
				if($row[$index] < $value){
					$newData[] = $row;
				}
			}else if($operator == '<'){
				if($row[$index] >= $value){
					$newData[] = $row;
				}
			}else if($operator == '<='){
				if($row[$index] > $value){
					$newData[] = $row;
				}
			}else if($operator == 'IN'){
				if(!in_array($row[$index], $value)){
					$newData[] = $row;
				}
			}else if($operator == 'LIKE'){
				if(!preg_match('/' . str_replace('%', '.*', $value) . '/', $row[$index])){
					$newData[] = $row;
				}
			}else{
				throw new Exception('Appel de Table::where(...) incorrect. Attendu : opérateur $operator valide (=, <>, >, IN, LIKE, etc.).');
			}
		}
		$this->data = $newData;
		return $this;
	}
	
	/**
	 * Retourne la table en format HTML: en-tête + données.
	 * @return String
	 **/
	public function html(){
		$html = '<table>';
		$html .= '	<thead>';
		$html .= '		<tr>';
		foreach($this->headers as $header){
			$html .= '		<th>' . $header . '</th>';
		}
		$html .= '		</tr>';
		$html .= '	</thead>';
		$html .= '	<tbody>';
		foreach($this->data as $row){
			$html .= '	<tr>';
			foreach($row as $column){
				// Pour afficher une table après regroupement(s) (i.e. après appel de la méthode Table::groupBy(...)).
				// En effet, en cas de regroupement(s), les lignes sont regroupées ou, en fait, plus exactement, les valeurs des colonnes de ces lignes sont regroupées : tableau. 
				if(is_array($column)){
					$html .= '<td>';
					foreach($column as $value){
						$html .= '<div>' . $value . '</div>';
					}
					$html .= '</td>';
				// En l'absence de regroupement(s), la valeur d'une colonne sur une ligne (une cellule quoi...) est unique. C'est une valeur scalaire : date, chaîne, caractère, entier, etc.
				}else{
					$html .= '<td>' . $column . '</td>';
				}
			}
			$html .= '	</tr>';
		}
		$html .= '	</tbody>';
		$html .= '</table>';
		return $html;
	}
	
}